"""EyeOnWater API integration."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .exceptions import EyeOnWaterException, EyeOnWaterResponseIsEmpty

if TYPE_CHECKING:
    from .client import Client
    from .meter_reader import MeterReader
    from .models import DataPoint, MeterInfo, Reading

    pass

SEARCH_ENDPOINT = "/api/2/residential/new_search"
CONSUMPTION_ENDPOINT = "/api/2/residential/consumption?eow=True"

MEASUREMENT_GALLONS = "GAL"
MEASUREMENT_100_GALLONS = "100 GAL"
MEASUREMENT_10_GALLONS = "10 GAL"
MEASUREMENT_CF = ["CF", "CUBIC_FEET"]
MEASUREMENT_CCF = "CCF"
MEASUREMENT_KILOGALLONS = "KGAL"
MEASUREMENT_CUBICMETERS = ["CM", "CUBIC_METER"]


_LOGGER = logging.getLogger(__name__)


class Meter:
    """Class represents meter state."""

    def __init__(self, reader: MeterReader) -> None:
        """Initialize the meter."""
        self.reader = reader
        self.last_historical_data: list[DataPoint] = []

        self._reading_data: Reading | None = None
        self._meter_info: MeterInfo | None = None

    @property
    def meter_uuid(self) -> str:
        """Return meter UUID."""
        return self.reader.meter_uuid

    @property
    def meter_id(self) -> str:
        """Return meter ID."""
        return self.reader.meter_id

    async def read_meter(self, client: Client, days_to_load: int = 3) -> None:
        """Triggers an on-demand meter read and returns it when complete."""
        self._meter_info = await self.reader.read_meter(client)
        self._reading_data = self._meter_info.reading

        try:
            # TODO: identify missing days and request only missing dates.
            historical_data = await self.reader.read_historical_data(
                days_to_load=days_to_load,
                client=client,
            )
            if not self.last_historical_data:
                self.last_historical_data = historical_data
            elif (
                historical_data
                and historical_data[-1].dt > self.last_historical_data[-1].dt
            ):
                # Take newer data
                self.last_historical_data = historical_data
            elif historical_data[-1].reading == self.last_historical_data[
                -1
            ].reading and len(historical_data) > len(self.last_historical_data):
                # If it the same date - take more data
                self.last_historical_data = historical_data

        except EyeOnWaterResponseIsEmpty:
            self.last_historical_data = []

    @property
    def meter_info(self) -> MeterInfo:
        """Return MeterInfo."""
        if not self._meter_info:
            msg = "Data was not fetched"
            raise EyeOnWaterException(msg)
        return self._meter_info

    @property
    def reading(self) -> float:
        """Returns the latest meter reading in me^3 or gal."""
        if not self._reading_data:
            msg = "Data was not fetched"
            raise EyeOnWaterException(msg)
        reading = self._reading_data.latest_read
        return self.reader.convert(reading.units, reading.full_read)



"""EyeOnWater API integration."""
from __future__ import annotations

import datetime
import json
import logging
from typing import TYPE_CHECKING, Any

from pydantic import ValidationError
import pytz

from .exceptions import EyeOnWaterAPIError, EyeOnWaterResponseIsEmpty
from .models import DataPoint, EOWUnits, HistoricalData, MeterInfo

if TYPE_CHECKING:
    from .client import Client

    pass

SEARCH_ENDPOINT = "/api/2/residential/new_search"
CONSUMPTION_ENDPOINT = "/api/2/residential/consumption?eow=True"

_LOGGER = logging.getLogger(__name__)


class MeterReader:
    """Class represents meter reader."""

    def __init__(
        self,
        meter_uuid: str,
        meter_info: dict[str, Any],
        metric_measurement_system: bool,
    ) -> None:
        """Initialize the meter."""
        self.meter_uuid = meter_uuid
        self.meter_id: str = meter_info["meter_id"]

        self.metric_measurement_system = metric_measurement_system
        self.native_unit_of_measurement = (
            "m\u00b3" if self.metric_measurement_system else "gal"
        )

    async def read_meter(self, client: Client) -> MeterInfo:
        """Triggers an on-demand meter read and returns it when complete."""
        _LOGGER.debug("Requesting meter reading")

        query = {"query": {"terms": {"meter.meter_uuid": [self.meter_uuid]}}}
        data = await client.request(path=SEARCH_ENDPOINT, method="post", json=query)
        data = json.loads(data)
        meters = data["elastic_results"]["hits"]["hits"]
        if len(meters) > 1:
            msg = "More than one meter reading found"
            raise Exception(msg)

        try:
            meter_info = MeterInfo.parse_obj(meters[0]["_source"])
        except ValidationError as e:
            msg = f"Unexpected EOW response {e}"
            raise EyeOnWaterAPIError(msg) from e

        return meter_info

    def convert(self, read_unit: EOWUnits, amount: float) -> float:
        """Convert reading to Cubic Meter or Gallons."""
        if self.metric_measurement_system:
            if read_unit in [EOWUnits.MEASUREMENT_CUBICMETERS, EOWUnits.MEASUREMENT_CM]:
                pass
            else:
                msg = f"Unsupported measurement unit: {read_unit}"
                raise EyeOnWaterAPIError(
                    msg,
                )
        else:
            if read_unit == EOWUnits.MEASUREMENT_KILOGALLONS:
                amount = amount * 1000
            elif read_unit == EOWUnits.MEASUREMENT_100_GALLONS:
                amount = amount * 100
            elif read_unit == EOWUnits.MEASUREMENT_10_GALLONS:
                amount = amount * 10
            elif read_unit == EOWUnits.MEASUREMENT_GALLONS:
                pass
            elif read_unit == EOWUnits.MEASUREMENT_CCF:
                amount = amount * 748.052
            elif read_unit in [
                EOWUnits.MEASUREMENT_CCF,
                EOWUnits.MEASUREMENT_CUBIC_FEET,
            ]:
                amount = amount * 7.48052
            else:
                msg = f"Unsupported measurement unit: {read_unit}"
                raise EyeOnWaterAPIError(
                    msg,
                )
        return amount

    async def read_historical_data(
        self,
        client: Client,
        days_to_load: int,
    ) -> list[DataPoint]:
        """Retrieve historical data for today and past N days."""
        today = datetime.datetime.now().replace(
            hour=0,
            minute=0,
            second=0,
            microsecond=0,
        )

        date_list = [today - datetime.timedelta(days=x) for x in range(0, days_to_load)]
        date_list.reverse()

        _LOGGER.info(
            f"requesting historical statistics for {self.meter_uuid} on {date_list}",
        )

        statistics = []

        for date in date_list:
            _LOGGER.info(
                f"requesting historical statistics for {self.meter_uuid} on {date}",
            )
            try:
                statistics += await self.read_historical_data_one_day(
                    date=date,
                    client=client,
                )
            except EyeOnWaterResponseIsEmpty:
                continue

        return statistics

    async def read_historical_data_one_day(
        self,
        client: Client,
        date: datetime.datetime,
    ) -> list[DataPoint]:
        """Retrieve the historical hourly water readings for a requested day."""
        if self.metric_measurement_system:
            units = "CM"
        else:
            units = self.native_unit_of_measurement.upper()

        query = {
            "params": {
                "source": "barnacle",
                "aggregate": "hourly",
                "units": units,
                "combine": "true",
                "perspective": "billing",
                "display_minutes": True,
                "display_hours": True,
                "display_days": True,
                "date": date.strftime("%m/%d/%Y"),
                "furthest_zoom": "hr",
                "display_weeks": True,
            },
            "query": {"query": {"terms": {"meter.meter_uuid": [self.meter_uuid]}}},
        }
        data = await client.request(
            path=CONSUMPTION_ENDPOINT,
            method="post",
            json=query,
        )
        try:
            data = HistoricalData.parse_raw(data)
        except ValidationError as e:
            msg = f"Unexpected EOW response {e}"
            raise EyeOnWaterAPIError(msg) from e

        key = f"{self.meter_uuid},0"
        if key not in data.timeseries:
            msg = f"Meter {key} not found"
            raise EyeOnWaterResponseIsEmpty(msg)

        timezones = data.hit.meter_timezone
        timezone = pytz.timezone(timezones[0])

        data = data.timeseries[key].series
        statistics = []
        for d in data:
            response_unit = d.display_unit.upper()
            statistics.append(
                DataPoint(
                    dt=timezone.localize(d.date),
                    reading=self.convert(response_unit, d.bill_read),
                ),
            )

        statistics.sort(key=lambda d: d.dt)

        return statistics



# ruff: noqa

from __future__ import annotations

from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field

from .units import EOWUnits


class Service(BaseModel):
    class_code_normalized: str
    route: str
    active: bool
    class_code: str
    service_type: str
    service_id: str
    start_date: datetime
    service_point_uuid: str
    cycle: str


class Location(BaseModel):
    city: str
    parcel_number: str
    location_name: str
    parity: str
    country: str
    route: str
    geocode_status: int
    zip_code: str
    longitude: str
    display_address_3: str
    state: str
    location_uuid: str
    county_name: str
    latitude: str
    display_address_2: str
    location_id: str
    display_address: str
    display_street_name: str
    cycle: str


class LeakAlert(BaseModel):
    alert_type: str
    name: str
    residential_user_name: str
    date_updated: Any
    alert_uuid: str
    state: str
    date_created: str
    creator_user_uuid: str


class Alerts(BaseModel):
    leak_alert: LeakAlert


class AccountInfo(BaseModel):
    status: str
    first_name: str
    billing_address: str
    billing_city: str
    account_uuid: str
    billing_address_2: str
    full_name: str
    email: str
    phone: str
    portal_status: str
    last_name: str
    account_id: str
    class_code: str
    billing_address_3: str
    person_id: str
    date_created: str
    account_billing_cycle: str
    billing_zip_code: str
    billing_country: str
    billing_state: str
    eyeonwater: str


class SensorsAvailable(BaseModel):
    types: list[str]


class Battery(BaseModel):
    register_: int = Field(..., alias="register")
    level: int
    quality: str
    thresh12mo: int
    time: datetime


class Pwr(BaseModel):
    level: int
    register_: int = Field(..., alias="register")
    signal_7days: int
    signal_strength: int
    time: datetime
    quality: str
    signal_30days: int


class Notes(BaseModel):
    count: int


class Flow(BaseModel):
    this_week: float
    months_updated: str
    last_month: float
    last_year_last_month_ratio: float
    last_year_last_month: float
    delta_positive: float
    time: datetime
    time_positive: str
    last_month_ratio: float
    last_week_avg: float
    last_year_this_month_ratio: float
    delta: float
    this_month: float
    week_ratio: float
    weeks_updated: str
    last_year_this_month: float
    last_week: float
    this_week_avg: float


class ActiveFlags(BaseModel):
    active_flags: list[str]
    time: datetime


class MeterData(BaseModel):
    sensors_available: SensorsAvailable
    has_endpoint: bool
    install_date: datetime
    meter_uuid: str
    fluid_type: str
    timezone: str
    firmware_version: str
    communication_security: str
    meter_size_unit: str
    cell_type: str
    geocode_status: int
    geo: str
    last_read_time: datetime
    note: str
    battery: Battery
    endpoint_status: str
    alert_code: int
    gas_pressure_compensation: float
    service_type: str
    serial_number: str
    type_: str = Field(..., alias="type")
    pwr: Pwr
    endpoint_connector: str
    communication_seconds: int
    meter_size_desc: str
    is_compound: str
    latitude: str
    typical_read_method: str
    endpoint_type: str
    meter_id: str
    last_communication_time: datetime
    pit_type: str
    meter_spec_uuid: str
    manufacturer: str
    has_valve: bool
    has_sensor: bool
    gas_sub_count: int
    notes: Notes
    meter_size: float
    flow: Flow
    longitude: str
    flags: ActiveFlags
    model: str
    sequence_number: int


class ServiceAgreement(BaseModel):
    service_agreement_uuid: str
    start_date: datetime


class LatestRead(BaseModel):
    bill_read: str
    bill_display_units: EOWUnits
    full_read: float
    read_time: datetime
    has_endpoints: bool
    units: EOWUnits
    method: str


class Timeslots(BaseModel):
    weekend: list[int]
    weekday: list[int]


class Encoder(BaseModel):
    time: datetime
    dials: int
    register_id: str
    totalizer: int


class Flags(BaseModel):
    forced: bool = Field(..., alias="Forced")
    magnetic_tamper: bool = Field(..., alias="MagneticTamper")
    empty_pipe: bool = Field(..., alias="EmptyPipe")
    leak: bool = Field(..., alias="Leak")
    encoder_no_usage: bool = Field(..., alias="EncoderNoUsage")
    encoder_temperature: bool = Field(..., alias="EncoderTemperature")
    encoder_reverse_flow: bool = Field(..., alias="EncoderReverseFlow")
    reading_changed: bool = Field(..., alias="ReadingChanged")
    cover_removed: bool = Field(..., alias="CoverRemoved")
    programming_changed: bool = Field(..., alias="ProgrammingChanged")
    encoder_exceeding_max_flow: bool = Field(..., alias="EncoderExceedingMaxFlow")
    water_temperature_sensor_error: bool = Field(
        ..., alias="WaterTemperatureSensorError"
    )
    oscillator_failure: bool = Field(..., alias="OscillatorFailure")
    encoder_sensor_error: bool = Field(..., alias="EncoderSensorError")
    tamper: bool = Field(..., alias="Tamper")
    reverse_flow: bool = Field(..., alias="ReverseFlow")
    encoder_leak: bool = Field(..., alias="EncoderLeak")
    low_battery: bool = Field(..., alias="LowBattery")
    water_pressure_sensor_error: bool = Field(..., alias="WaterPressureSensorError")
    min_max_invalid: bool = Field(..., alias="MinMaxInvalid")
    end_of_life: bool = Field(..., alias="EndOfLife")
    encoder_dial_change: bool = Field(..., alias="EncoderDialChange")
    no_usage: bool = Field(..., alias="NoUsage")
    battery_charging: bool = Field(..., alias="BatteryCharging")
    device_alert: bool = Field(..., alias="DeviceAlert")
    endpoint_reading_missed: bool = Field(..., alias="EndpointReadingMissed")
    encoder_removal: bool = Field(..., alias="EncoderRemoval")
    profile_read_error: bool = Field(..., alias="ProfileReadError")
    encoder_programmed: bool = Field(..., alias="EncoderProgrammed")
    time: datetime
    encoder_magnetic_tamper: bool = Field(..., alias="EncoderMagneticTamper")
    meter_temperature_sensor_error: bool = Field(
        ..., alias="MeterTemperatureSensorError"
    )


class Reading(BaseModel):
    battery: Battery
    customer_uuid: str
    aggregation_seconds: int
    last_communication_time: datetime
    firmware_version: str
    communication_security: str
    meter_size_desc: str
    barnacle_uuid: str
    unit: EOWUnits
    customer_name: str
    cell_type: str
    pi_status: str
    second_carrier: bool
    latest_read: LatestRead
    input_config: str
    meter_size_unit: str
    endpoint_status: str
    gas_pressure_compensation: float
    serial_number: str
    activated_on: str
    sim_vendor: str
    pwr: Pwr
    communication_seconds: int
    cell_endpoint_name: str
    rf_communication: bool
    low_read_limit: int
    utility_use_1: str
    timeslots: Timeslots
    encoder: Encoder
    endpoint_type: str
    utility_use_2: str
    multiplier: str
    sim_type: str
    register_number: str
    wired_interface: str
    endpoint_install_date: datetime
    high_read_limit: int
    gas_sub_count: int
    billing_number: str
    meter_size: float
    flow: Flow
    hardware_version: str
    connector_type: str
    flags: Flags
    model: str
    resolution: float


class Conditions(BaseModel):
    increasing: bool
    decreasing: bool


class EndpointTemperature(BaseModel):
    latest_average: float
    last_reported: str
    seven_day_min: str
    seven_day_average: str
    seven_day_max: str
    sensor_uuid: str
    conditions: Conditions


class Sensors(BaseModel):
    endpoint_temperature: EndpointTemperature


class Utility(BaseModel):
    fluid_barrel_billing_unit: str
    cm_billing_unit: str
    cf_billing_unit: str
    type_: str = Field(..., alias="type")
    utility_name: str
    eow_service_selector: str
    gas_cf_billing_unit: str
    eow_type: str
    oil_barrel_billing_unit: str
    date_created: str
    gal_billing_unit: str
    gas_cm_billing_unit: str
    utility_uuid: str
    imp_billing_unit: str


class User(BaseModel):
    user_uuid: str
    user_name: str
    date_created: str


class Groups(BaseModel):
    irrigation: str
    continuous_flow: str
    is_irrigatable: str
    disable_valve_shutoff: str


class MeterInfo(BaseModel):
    reading: Reading = Field(..., alias="register_0")
    sensors: Sensors
    utility: Utility
    updated: int
    last_updated: str
    service: Service
    location: Location
    alerts: Alerts
    account: AccountInfo
    meter: MeterData
    service_agreement: ServiceAgreement
    version: str
    user: User
    groups: Groups



"""Supported EOW units."""

from __future__ import annotations

from enum import Enum


class EOWUnits(str, Enum):
    """Enum of supported EOW units."""

    MEASUREMENT_GALLONS = "GAL"
    MEASUREMENT_100_GALLONS = "100 GAL"
    MEASUREMENT_10_GALLONS = "10 GAL"
    MEASUREMENT_CF = "CF"
    MEASUREMENT_CUBIC_FEET = "CUBIC_FEET"
    MEASUREMENT_CCF = "CCF"
    MEASUREMENT_KILOGALLONS = "KGAL"
    MEASUREMENT_CM = "CM"
    MEASUREMENT_CUBICMETERS = "CUBIC_METER"



# ruff: noqa

from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel, Field

from .units import EOWUnits


class Register0EncoderItem(BaseModel):
    dials: int


class Hit(BaseModel):
    meter_communication_seconds: list[int] = Field(
        ..., alias="meter.communication_seconds"
    )
    meter_timezone: list[str] = Field(..., alias="meter.timezone")
    register_0_encoder: list[Register0EncoderItem] = Field(
        ..., alias="register_0.encoder"
    )
    location_location_uuid: list[str] = Field(..., alias="location.location_uuid")
    meter_fluid_type: list[str] = Field(..., alias="meter.fluid_type")
    meter_meter_id: list[str] = Field(..., alias="meter.meter_id")
    account_full_name: list[str] = Field(..., alias="account.full_name")
    meter_meter_uuid: list[str] = Field(..., alias="meter.meter_uuid")
    meter_has_endpoint: list[bool] = Field(..., alias="meter.has_endpoint")
    meter_serial_number: list[str] = Field(..., alias="meter.serial_number")
    account_account_id: list[str] = Field(..., alias="account.account_id")
    location_location_name: list[str] = Field(..., alias="location.location_name")
    service_service_id: list[str] = Field(..., alias="service.service_id")
    register_0_serial_number: list[str] = Field(..., alias="register_0.serial_number")
    utility_utility_uuid: list[str] = Field(..., alias="utility.utility_uuid")
    account_account_uuid: list[str] = Field(..., alias="account.account_uuid")


class Params(BaseModel):
    start_date_utc: float
    date: datetime
    end_date: datetime
    end_date_utc: float
    compare: bool
    read_type: str
    start_date: datetime
    end_date_tz: str
    aggregate: str
    aggregate_group: str
    perspective: str
    units: EOWUnits
    start_date_tz: str
    adjust_to: bool
    combine_group: bool


class Series(BaseModel):
    bill_read: float
    end_date: datetime
    display_unit: EOWUnits
    meter_uuid: int
    value: float
    start_date: datetime
    date: datetime
    register_number: int
    estimated: int
    raw_read: int
    unit: EOWUnits


class Legend(BaseModel):
    supply_zone_id: str
    location_name: str
    account_id: str
    demand_zone_id: str
    meter_id: str
    full_name: str
    serial_number: str


class TimeSerie(BaseModel):
    series: list[Series]
    legend: Legend


class HistoricalData(BaseModel):
    hit: Hit
    min_chart_aggregation: str
    params: Params
    timeseries: dict[str, TimeSerie]
    timezone: str
    min_aggregation_seconds: int
    annotations: list[str]
